Fixed issue with focusing on minimized split-screen stack
If the current split-screen secondary stack is finishing and we are
moving focus to another stack, we should prefer the focus to move to
another split-screen secondary stack or fullscreen stack instead of
the primary split screen stack to avoid:
- Moving the focus to the primary split-screen stack when it can't
be focused because it will be minimized, but AM doesn't know that yet
- primary split-screen stack overlapping with a fullscreen stack when
a fullscreen stack is higher in z than the next split-screen stack.
Assistant stack, I am looking at you...
We only move the focus to the primary-split screen stack if there isn't
a better alternative.
Change-Id: Iffaaa56a88e24bbf0f6303acd924700257baad0b
Fixes: 74404706
Fixes: 70677280
Fixes: 74366688
Test: atest ActivityManagerAssistantStackTests#testAssistantStackLaunchNewTaskWithDockedStack
diff --git a/services/core/java/com/android/server/am/ActivityStack.java b/services/core/java/com/android/server/am/ActivityStack.java
index 7749bd7..38b17bb 100644
--- a/services/core/java/com/android/server/am/ActivityStack.java
+++ b/services/core/java/com/android/server/am/ActivityStack.java
@@ -3422,8 +3422,8 @@
* @param allowFocusSelf Is the focus allowed to remain on the same stack.
*/
private boolean adjustFocusToNextFocusableStack(String reason, boolean allowFocusSelf) {
- final ActivityStack stack = mStackSupervisor.getNextFocusableStackLocked(
- allowFocusSelf ? null : this);
+ final ActivityStack stack =
+ mStackSupervisor.getNextFocusableStackLocked(this, !allowFocusSelf);
final String myReason = reason + " adjustFocusToNextFocusableStack";
if (stack == null) {
return false;
diff --git a/services/tests/servicestests/src/com/android/server/am/ActivityTestsBase.java b/services/tests/servicestests/src/com/android/server/am/ActivityTestsBase.java
index 1195188..4a19345 100644
--- a/services/tests/servicestests/src/com/android/server/am/ActivityTestsBase.java
+++ b/services/tests/servicestests/src/com/android/server/am/ActivityTestsBase.java
@@ -371,7 +371,8 @@
// Just return the current front task. This is called internally so we cannot use spy to mock this out.
@Override
- ActivityStack getNextFocusableStackLocked(ActivityStack currentFocus) {
+ ActivityStack getNextFocusableStackLocked(ActivityStack currentFocus,
+ boolean ignoreCurrent) {
return mFocusedStack;
}
}
diff --git a/services/core/java/com/android/server/am/ActivityStackSupervisor.java b/services/core/java/com/android/server/am/ActivityStackSupervisor.java
index 0216439..869c635 100644
--- a/services/core/java/com/android/server/am/ActivityStackSupervisor.java
+++ b/services/core/java/com/android/server/am/ActivityStackSupervisor.java
@@ -679,7 +679,7 @@
void setFocusStackUnchecked(String reason, ActivityStack focusCandidate) {
if (!focusCandidate.isFocusable()) {
// The focus candidate isn't focusable. Move focus to the top stack that is focusable.
- focusCandidate = getNextFocusableStackLocked(focusCandidate);
+ focusCandidate = getNextFocusableStackLocked(focusCandidate, false /* ignoreCurrent */);
}
if (focusCandidate != mFocusedStack) {
@@ -2429,27 +2429,50 @@
* Get next focusable stack in the system. This will search across displays and stacks
* in last-focused order for a focusable and visible stack, different from the target stack.
*
* @param currentFocus The stack that previously had focus and thus needs to be ignored when
* searching for next candidate.
* @return Next focusable {@link ActivityStack}, null if not found.
*/
- ActivityStack getNextFocusableStackLocked(ActivityStack currentFocus) {
+ ActivityStack getNextFocusableStackLocked(ActivityStack currentFocus, boolean ignoreCurrent) {
mWindowManager.getDisplaysInFocusOrder(mTmpOrderedDisplayIds);
+ final int currentWindowingMode = currentFocus != null
+ ? currentFocus.getWindowingMode() : WINDOWING_MODE_UNDEFINED;
+ ActivityStack candidate = null;
for (int i = mTmpOrderedDisplayIds.size() - 1; i >= 0; --i) {
final int displayId = mTmpOrderedDisplayIds.get(i);
// If a display is registered in WM, it must also be available in AM.
final ActivityDisplay display = getActivityDisplayOrCreateLocked(displayId);
for (int j = display.getChildCount() - 1; j >= 0; --j) {
final ActivityStack stack = display.getChildAt(j);
- if (stack != currentFocus && stack.isFocusable()
- && stack.shouldBeVisible(null)) {
- return stack;
+ if (ignoreCurrent && stack == currentFocus) {
+ continue;
}
+ if (!stack.isFocusable() || !stack.shouldBeVisible(null)) {
+ continue;
+ }
+
+ if (currentWindowingMode == WINDOWING_MODE_SPLIT_SCREEN_SECONDARY
+ && candidate == null && stack.inSplitScreenPrimaryWindowingMode()) {
+ // If the currently focused stack is in split-screen secondary we would prefer
+ // the focus to move to another split-screen secondary stack or fullscreen stack
+ // over the primary split screen stack to avoid:
+ // - Moving the focus to the primary split-screen stack when it can't be focused
+ // because it will be minimized, but AM doesn't know that yet
+ // - primary split-screen stack overlapping with a fullscreen stack when a
+ // fullscreen stack is higher in z than the next split-screen stack. Assistant
+ // stack, I am looking at you...
+ // We only move the focus to the primary-split screen stack if there isn't a
+ // better alternative.
+ candidate = stack;
+ continue;
+ }
+ return stack;
}
}
- return null;
+ return candidate;
}
/**
@@ -2429,27 +2429,50 @@
* in last-focused order for a focusable and visible stack, different from the target stack.
*
- * @param currentFocus The stack that previously had focus and thus needs to be ignored when
- * searching for next candidate.
+ * @param currentFocus The stack that previously had focus.
+ * @param ignoreCurrent If we should ignore {@param currentFocus} when searching for next
+ * candidate.
* @return Next focusable {@link ActivityStack}, null if not found.
*/